#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import pickle as pkl

from .boruta import Boruta
from .naive import NaiveLearner
from .rfe import RFE


class BasePredictor(object):
	def _predict(self, obj):
		assert hasattr(self, 'models'), 'The model should be first fitted'
		assert hasattr(self, 'selected_variables'), 'The model should be first fitted'
		assert len(self.selected_variables) > 0, 'There should be at least one feature selected'

		x = obj[self.selected_variables].values
		y = self.models[0].predict(x)
		predictions = pd.DataFrame(index=obj.index)
		predictions[self.target_column] = y

		return predictions


	def predict(self, obj, memory_bound=False):
		"""
		Make predictions using the fitted model.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing test explanatory variables/features about which we want to make predictions.
		memory_bound : bool (Default False)
			Whether we should try to save memory.


		Returns
		-------
		result : pandas.DataFrame
			A dataframe with the same index as :code:`obj`, and with one column whose name is the :code:`target_column` used for training.
		"""
		if memory_bound:
			n = obj.shape[0]
			max_n = 1000000
			res = pd.DataFrame(index=obj.index)
			res[self.target_column] = np.nan
			i = 0
			while i < n:
				res.iloc[i:i+max_n] = self._predict(obj.iloc[i:i+max_n])
				i += max_n
			return res

		else:
			return self._predict(obj)


	def save(self, path):
		"""
		Cache the predictor to disk.
		"""
		meta_path = path + '-meta-' + self.__class__.__name__
		meta = {'target_column': self.target_column, 'selected_variables': self.selected_variables}
		with open(meta_path, 'wb') as f:
			pkl.dump(meta, f)
		self.models[0].save(path + '-' + self.__class__.__name__)
		

	@classmethod
	def load(cls, path, learner_func):
		"""
		Load the predictor from disk.
		"""
		meta_path = path + '-meta-' + cls.__name__
		with open(meta_path, 'rb') as f:
			meta = pkl.load(f)
		target_column = meta['target_column']
		selected_variables = meta['selected_variables']

		n_vars = len(selected_variables)
		model = learner_func(n_vars=n_vars, path=path + '-' + cls.__name__)
		
		predictor = cls()
		predictor.models = [model]
		predictor.selected_variables = selected_variables
		predictor.target_column = target_column

		return predictor




class RFEPredictor(BasePredictor):
	def fit(self, obj, target_column, learner_func, n_features, max_duration=None):
		"""
		Fits a supervised learner enriched with feature selection using the Recursive Feature Elimination algorithm.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing training explanatory variables/features as well as the target.
		target_column : str
			The name of the column in :code:`obj` containing targets.
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features. Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained. There should be as many importance scores in :code:`feature_importances_` as columns in :code:`fit(x_train, y_train)`.
		n_features : int
			The number of features to keep.
		max_duration : float | None (default)
			If not None, then feature elimination will stop after this many seconds.


		Attributes
		----------
		selected_variables : list
			The list of the :code:`n_features` features we kept.
		target_column : str
			The name of the column used as target.
		models : list
			An array whose first entry is the fitted model.


		Returns
		-------
		results : dict
			A dictionary containing, among other things, selected variables.

		"""
		self.target_column = target_column
		x_columns = [_ for _ in obj.columns if _ != target_column]
		x_df = obj[x_columns]
		y_df = obj[[target_column]]

		feature_selector = RFE(learner_func)
		m = feature_selector.fit(x_df, y_df, n_features, max_duration=max_duration)
		self.models = [m]
		self.selected_variables = feature_selector.selected_variables

		results = {'Selected Variables': self.selected_variables}
		return results



class BorutaPredictor(BasePredictor):
	def fit(self, obj, target_column, learner_func, n_evaluations=20, pval=0.95, max_duration=None):
		"""
		Fits a supervised learner enriched with feature selection using the Boruta algorithm.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing training explanatory variables/features as well as the target.
		target_column : str
			The name of the column in :code:`obj` containing targets.
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features. Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained. There should be as many importance scores in :code:`feature_importances_` as columns in :code:`fit(x_train, y_train)`.
		n_evaluations : int (Default 20)
			The number of trials to run in the Boruta algorithm.
		pval : float (Default 0.95)
			The quantile level above which to consider a feature relevant in the Boruta algorithm.
		max_duration : float | None (default)
			If not None, then feature elimination will stop after this many seconds.

		Attributes
		----------
		selected_variables : list
			The list of the :code:`n_features` features we kept.
		target_column : str
			The name of the column used as target.
		models : list
			An array whose first entry is the fitted model.


		Returns
		-------
		results : dict
			A dictionary containing, among other things, selected variables.

		"""
		self.target_column = target_column
		x_columns = [_ for _ in obj.columns if _ != target_column]
		x_df = obj[x_columns]
		y_df = obj[[target_column]]

		feature_selector = Boruta(learner_func)
		m = feature_selector.fit(x_df, y_df, n_evaluations=n_evaluations, pval=pval, max_duration=max_duration)
		self.models = [m]
		self.selected_variables = feature_selector.selected_variables

		results = {'Selected Variables': self.selected_variables}
		return results



class NaivePredictor(BasePredictor):
	def fit(self, obj, target_column, learner_func):
		"""
		Fits a supervised learner without feature selection.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing training explanatory variables/features as well as the target.
		target_column : str
			The name of the column in :code:`obj` containing targets.
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features. Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained. There should be as many importance scores in :code:`feature_importances_` as columns in :code:`fit(x_train, y_train)`.

		Attributes
		----------
		selected_variables : list
			The list of the :code:`n_features` features we kept.
		target_column : str
			The name of the column used as target.
		models : list
			An array whose first entry is the fitted model.


		Returns
		-------
		results : dict
			A dictionary containing, among other things, selected variables.

		"""
		self.target_column = target_column
		x_columns = [_ for _ in obj.columns if _ != target_column]
		x_df = obj[x_columns]
		y_df = obj[[target_column]]

		feature_selector = NaiveLearner(learner_func)
		m = feature_selector.fit(x_df, y_df)
		self.models = [m]
		self.selected_variables = feature_selector.selected_variables

		results = {'Selected Variables': self.selected_variables}
		return results











#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Python client for the KXY RESTful API.
"""

from functools import lru_cache
import os
import requests

from .decorators import requires_api_key, get_api_key
from .. import __version__ as client_version


class APIClient(object):
	"""
	Python client for the RESTful KXY API. All API methods require an API key. The API key must be set by running :code:`kxy configure` from the terminal.
	"""
	@staticmethod
	def stage():
		"""
		Defines the deployment stage of the RESTful API the client should talk to.

		Returns
		-------
		v : str
			The API stage to use.
		"""
		return 'v2'

	@staticmethod
	def url(path):
		"""
		Turns a relative path into a full API endpoint url.

		Parameters
		----------
		path: str
			The relative path of the API resource.

		Returns
		-------
		u : str
			The full URL of the API resource.
		"""
		path = path.strip('/')

		return 'https://api.kxy.ai/%s/' % APIClient.stage() + path


	@staticmethod
	@requires_api_key
	def get(path, **params):
		"""
		.. important:: This method requires a valid API key.

		Issues a GET request to the API resource identified by the input path.

		Parameters
		----------
		path: str
			The relative path of the API resource.
		params: dict, optional
			The query parameters of the GET request. Any keyword argument is 
			automatically interpreted as a request parameter, its name is used
			as the parameter name, and its value as the parameter value.

		Returns
		-------
		response: requests.Response
			The response of the API. The request HTTP status code can be accessed
			through `response.status_code`. To check if the request was succesful,
			inspect `response.ok`. When the API returned data, they can be accessed
			through `response.json()`. Supported status codes are:

			200: 
				The request was successful and the API returned some data accessible through
				`response.json()`.
			402: 
				The request failed because your account does not have a valid payment method.
				Check `response.json()['reason']` for more information.
			403: 
				The request failed because some parameter are either invalid or missing.
				Check `response.json()['reason']` for more information.
			404:
				The request failed because the API couldn't yet solve the problem of interest.
				You should typically try again another time. Check `response.json()['reason']`
				for more information.
		"""
		url = APIClient.url(path)
		api_key = get_api_key()
		if 'client_version' not in params:
			params['client_version'] = client_version
		response = requests.get(url, params=params, headers={'x-api-key': api_key, \
			'content-type': 'application/json'})

		return response


	@staticmethod
	@requires_api_key
	def post(path, **params):
		"""
		.. important:: This method requires a valid API key.

		Issues a POST request to the API resource identified by the input path.

		Parameters
		----------
		path: str
			The relative path of the API resource.
		params: dict, optional
			The data to be submitted to the API as part of the POST request, as 
			a JSON. Any keyword argument is automatically interpreted as a 
			key of the JSON data that will be submitted to the API, 
			and its value the associated value in the JSON.

		Returns
		-------
		response: requests.Response
			The response of the API. The request HTTP status code can be accessed
			through `response.status_code`. To check if the request was succesful,
			inspect `response.ok`. When the API returned data, they can be accessed
			through `response.json()`.

			Supported status codes are:

			200: 
				The request was successful and the API returned some data accessible through
				`response.json()`.
			402: 
				The request failed because your account does not have a valid payment method.
				Check `response.json()['reason']` for more information.
			403: 
				The request failed because some parameter are either invalid or missing.
				Check `response.json()['reason']` for more information.
			404:
				The request failed because the API couldn't yet solve the problem of interest.
				You should typically try again another time. Check `response.json()['reason']`
				for more information.
		"""
		url = APIClient.url(path)
		api_key = get_api_key()
		if 'client_version' not in params:
			params['client_version'] = client_version
		response = requests.post(url, json=params, headers={'x-api-key': api_key, \
			'content-type': 'application/json'})

		return response


	@staticmethod
	@lru_cache(maxsize=32)
	def route(path=None, method=None, **params):
		"""
		.. important:: This method requires a valid API key.

		Generic method to issue a GET or a POST request to the API resource identified
		by the input path.

		Parameters
		----------
		path: str
			The relative path of the API resource.

		method: str
			The REST method. Should be either `'GET'` or `'POST'`.

		params: dict, optional
			The data to be submitted to the API as a JSON for POST requests, or
			query parameters in the case of GET requests.

		Returns
		-------
		response: requests.Response
			The response of the API. The request HTTP status code can be accessed
			through `response.status_code`. To check if the request was succesful,
			inspect `response.ok`. When the API returned data, they can be accessed
			through `response.json()`.

			Supported status codes are:

			200: 
				The request was successful and the API returned some data accessible through
				`response.json()`.
			402: 
				The request failed because your account does not have a valid payment method.
				Check `response.json()['reason']` for more information.
			403: 
				The request failed because some parameter are either invalid or missing.
				Check `response.json()['reason']` for more information.
			404:
				The request failed because the API couldn't yet solve the problem of interest.
				You should typically try again another time. Check `response.json()['reason']`
				for more information.

		Raises
		------
		ValueError
			If path is None or method is neither 'GET', nor 'POST'.
		"""
		if path is None or method is None or \
				method.upper() not in ('GET', 'POST'):
			return None

		if method.upper() == 'GET':
			return APIClient.get(path, **params)

		if method.upper() == 'POST':
			return APIClient.post(path, **params)







#!/usr/bin/env python
# -*- coding: utf-8 -*-
import hashlib
import numpy as np
from scipy.stats import norm
import pandas as pd


class BaseAccessor(object):
	"""
	Base class inheritated by our customs accessors.
	"""
	def __init__(self, pandas_obj):
		self._obj = pandas_obj


	def is_discrete(self, column):
		"""
		Determine whether the input column contains discrete (i.e as opposed to continuous) observations.
		"""
		if self.is_categorical(column):
			return True

		n = self._obj.shape[0]
		values, counts = np.unique(self._obj[column].values, return_counts=True)
		unique_n = len(values)

		if unique_n < 0.05*n:
			return True

		counts = np.array(list(sorted(counts)))
		if np.sum(counts[-10:]) > 0.8*n:
			return True

		return False


	def is_categorical(self, column):
		"""
		Determine whether the input column contains categorical (i.e. non-ordinal) observations.
		"""
		if self._obj[column].dtype in [float, int, np.float32, np.float64, np.int32, np.int64]:
			return False
			
		try:
			casted = self._obj[column].values.astype(float)
			return False
		except:
			return True


	@property
	def is_too_large(self):
		return self._obj.memory_usage(index=False).sum()/(1024.0*1024.0*1024.0) > 1.5


	def describe(self,):
		for col in sorted(self._obj.columns):
			print('         ')
			print('---------' + '-'.join(['' for c in col]))
			print('Column: %s' % col)
			print('---------' + '-'.join(['' for c in col]))
			if self._obj.kxy.is_categorical(col):
				print('Type:      Categorical')
				labels, counts = np.unique(self._obj[col].values.astype(str), return_counts=True)
				labels_with_counts = [(labels[i], 100.*counts[i]/self._obj.shape[0]) \
									  for i in range(len(labels))]
				labels_with_counts = sorted(labels_with_counts, key=lambda x: -x[1])
				tot = 0.0
				for label, freq in labels_with_counts:
					print('Frequency: %s%%, Label: %s' % (('%.2f' % freq).rjust(5, ' '), label))
					tot += freq
					if tot > 90. and tot < 100.:
						print('Other Labels: %.2f%%' % (100.-tot))
						break
			else:
				m   = self._obj[col].min(skipna=True, numeric_only=True)
				M   = self._obj[col].max(skipna=True, numeric_only=True)
				mn  = self._obj[col].mean(skipna=True, numeric_only=True)
				q50 = self._obj[col].median(skipna=True, numeric_only=True)
				q25 = self._obj[col].quantile(0.25, numeric_only=True)
				q75 = self._obj[col].quantile(0.75, numeric_only=True)

				print('Type:   Continuous')
				print('Max:    %s' % ('%.1f' % M if M < 10. else '{:,}'.format(int(M))))
				print('p75:    %s' % ('%.1f' % q75 if q75 < 10. else '{:,}'.format(int(q75))))
				print('Mean:   %s' % ('%.1f' % mn if mn < 10. else '{:,}'.format(int(mn))))
				print('Median: %s' % ('%.1f' % q50 if q50 < 10. else '{:,}'.format(int(q50))))
				print('p25:    %s' % ('%.1f' % q25 if q25 < 10. else '{:,}'.format(int(q25))))
				print('Min:    %s' % ('%.1f' % m if m < 10. else '{:,}'.format(int(m))))


	def anonymize(self, columns_to_exclude=[]):
		"""
		Anonymize the dataframe in a manner that leaves all pre-learning and post-learning analyses (including data valuation, variable selection, model-driven improvability, data-driven improvability and model explanation) invariant.

		Any transformation on continuous variables that preserves ranks will not change our pre-learning and post-learning analyses. The same holds for any 1-to-1 transformation on categorical variables.

		This implementation replaces ordinal values (i.e. any column that can be cast as a float) with their within-column Gaussian score. For each non-ordinal column, we form the set of all possible values, we assign a unique integer index to each value in the set, and we systematically replace said value appearing in the dataframe by the hexadecimal code of its associated integer index. 

		For regression problems, accurate estimation of RMSE related metrics require the target column (and the prediction column for post-learning analyses) not to be anonymized.


		Parameters
		----------
		columns_to_exclude: list (optional)
			List of columns not to anonymize (e.g. target and prediction columns for regression problems).


		Returns
		-------
		result : pandas.DataFrame
			The result is a pandas.Dataframe with columns (where applicable):
		"""
		df = self._obj.copy()
		for col in df.columns:
			if col in columns_to_exclude:
				continue

			if df.kxy.is_categorical(col):
				unique_values = list(sorted(set(list(df[col].values))))
				mapping = {unique_values[i]: "0x{:03x}".format(i) for i in range(len(unique_values))}
				df[col] = df[col].apply(lambda x: mapping.get(x))
			else:
				# Note: Any monotonic transformation applied to any continuous column would work.
				# The gaussian scoring below makes no assumption on marginals whatsoever. 
				x = df[col].values
				x = x - np.nanmean(x)
				s = np.nanstd(x)
				if s > 0.0:
					x = x/s
					x = norm.cdf(x)
				df[col] = np.around(x.copy(), 3)


		return df



	def __hash__(self):
		return hashlib.sha256(self._obj.to_string().encode()).hexdigest()







#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd

from .base_accessor import BaseAccessor
from ..learning.shrunk_learner import ShrunkLearner as LeanMLPredictor
from ..misc.predictors import BorutaPredictor, RFEPredictor, NaivePredictor



@pd.api.extensions.register_dataframe_accessor("kxy_learning")
class LearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for automatically training predictive models.

	This class defines the :code:`kxy_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""


	def fit(self, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
			max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=3, \
			regression_baseline='mean', additive_learning=False, regression_error_type='additive', return_scores=False, \
			start_n_features_perf_frac=0.9, feature_selection_method='leanml', rfe_n_features=None, boruta_pval=0.5, \
			boruta_n_evaluations=20, max_duration=None, val_performance_buffer=0.0):
		"""
		Train a lean boosted supervised learner, bringing in variables one at a time, in decreasing order of importance (as per :code:`df.kxy.variable_selection`), until doing so no longer improves validation performance or another stopping criterion is met.

		Specifically, training proceeds as follows. First, KXY's model-free variable selection is run (i.e. :code:`df.kxy.variable_selection`). 

		Then we train a model (instance returned by :code:`learner_func`) using the :code:`start_n_features` most important feature/variable to predict the target (defined by :code:`target_column`).

		When :code:`start_n_features` is :code:`None` the initial set of variables is the smallest set of variables with which we may achieve :code:`start_n_features_perf_frac` of the performance we could achieve using all variables (as per :code:`df.kxy.variable_selection`).

		Next we consider adding one variable at a time to fix the mistakes made by the previously trained model when :code:`additive_learning` is :code:`True`.
	
		If doing so improves performance on the validation set, we keep going until either performance no longer improves on the validation set :code:`n_down_perf_before_stop` consecutive times, or we've selected :code:`max_n_features` features.

		When :code:`additive_learning` is set to :code:`False` (the default), after adding a new variable, we train the new model on the original problem, rather than trying to improve residuals.


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		learner_func : function
			A function returning an instance of a base learner. They should define a :code:`fit(x, y)` method and a :code:`predict(x)` method.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		snr : 'auto' | 'low' | 'high'
			Set to :code:`low` if the problem is difficult (i.e. has a low signal-to-noise ratio) or the number of rows is small relative to the number of columns. Only used for model-free variable selection.
		train_frac : float
			The fraction of rows used for training and validation.
		random_state : int
			The seed to use for random training/validation/testing split.
		min_n_features : int | None
			Boosting will not stop until at least this many features/explanatory variables are selected.
		max_n_features : int | None
			Boosting will stop as soon as this many features/explanatory variables are selected.
		start_n_features : int
			The number of most important features boosting will start with.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).
		benchmark_feature : str | None
			When not None, 'benchmark' performance metrics using this column as predictor will be reported in the output dictionary.
		missing_value_imputation : bool
			When set to True, replace missing values with medians.
		n_down_perf_before_stop : int
			Number of consecutive down performances to observe before boosting stops.
		regression_baseline : str (:code:`mean` | :code:`median`)
			Whether to use the unconditional mean or median as the best predictor in the absence of explanatory variables.
			Choosing the mean corresponds to minimizing the L2 norm, whereas choosing the median corresponds to minimizing the L1 norm.
		additive_learning : bool
			When a new variable is added, whether errors/residuals should be fixed or a new model should be learned from scratch.
		regression_error_type : str ('additive' | 'multiplicative')
			For regression problems with additive learning, this determines whether the final model should be additive (pruning tries to reduce regressor residuals) or multiplicative (i.e. pruning tries to bring the ratio between true and predicted labels as closed to 1 as possible).
		start_n_features_perf_frac : float (between 0 and 1)
			When :code:`start_n_features` is not specified, it is set to the number of variables required to achieve a fraction :code:`start_n_features_perf_frac` of the maximum performance achievable (as per :code:`df.kxy.variable_selection`).
		return_scores : bool (Default False)
			Whether to return training, validation and testing performance after lean boosting.
		feature_selection_method : str (:code:`leanml` | :code:`rfe` | :code:`boruta` | :code:`none`. Default :code:`leanml`)
			Do not change this unless you want to try out Boruta or Recursive Feature Selection. The leanml method outperforms both.
		rfe_n_features : int
			The number of features to keep when the feature selection method is :code:`rfe`.
		boruta_pval : float
			The quantile level to use when the feature selection method is :code:`boruta`.
		boruta_n_evaluations : int
			The number of trials to use when the feature selection method is :code:`boruta`.
		max_duration : float | None (default)
			If not None, then Boruta and RFE will stop after this many seconds.
		val_performance_buffer : float (Default 0.0)
			In LeanML feature selection, this is the threshold by which the new validation performance needs to exceed the previously evaluated validation performance to consider increasing the number of features.


		Returns
		-------
		result : dict
			Dictionary containing selected variables, as well as training, validation and testing performance, and the trained model.
		"""

		if str(feature_selection_method).lower() == 'none':
			predictor = NaivePredictor()
			res = predictor.fit(self._obj, target_column, learner_func)
			self.predictor = predictor
			res['predictor'] = predictor

		elif feature_selection_method.lower() == 'leanml':
			predictor = LeanMLPredictor()
			res = predictor.fit(self._obj, target_column, learner_func, problem_type=problem_type, snr=snr, train_frac=train_frac, random_state=random_state, \
					max_n_features=max_n_features, min_n_features=min_n_features, start_n_features=start_n_features, anonymize=anonymize, \
					benchmark_feature=benchmark_feature, missing_value_imputation=missing_value_imputation, score=score, n_down_perf_before_stop=n_down_perf_before_stop, \
					regression_baseline=regression_baseline, regression_error_type=regression_error_type, return_scores=return_scores, \
					start_n_features_perf_frac=start_n_features_perf_frac, val_performance_buffer=val_performance_buffer)
			self.predictor = predictor
			res['predictor'] = predictor

		elif feature_selection_method.lower() == 'boruta':
			predictor = BorutaPredictor()
			res = predictor.fit(self._obj, target_column, learner_func, pval=boruta_pval, n_evaluations=boruta_n_evaluations, max_duration=max_duration)
			self.predictor = predictor
			res['predictor'] = predictor

		elif feature_selection_method.lower() == 'rfe':
			assert rfe_n_features is not None
			predictor = RFEPredictor()
			res = predictor.fit(self._obj, target_column, learner_func, n_features=rfe_n_features, max_duration=max_duration)
			self.predictor = predictor
			res['predictor'] = predictor

		else:
			raise ValueError('The value of feature_selection_method (%s) is not allowed' % feature_selection_method)


		return res


	def predict(self, obj, memory_bound=False):
		"""
		Make predictions using the fitted model.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing test explanatory variables about which we want to make predictions.
		memory_bound : bool (Default False)
			Whether we should try to save memory.


		Returns
		-------
		result : pandas.DataFrame
			A dataframe with the same index as :code:`obj`, and with one column whose name is the :code:`target_column` used for training.
		"""
		assert hasattr(self, 'predictor'), 'The model should be first fitted'
		return self.predictor.predict(obj, memory_bound=memory_bound)








#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from time import time
import numpy as np
import pandas as pd
from scipy.stats import binom

from tqdm import tqdm

class Boruta(object):
	"""
	Implementation of the Boruta feature selection algorithm.

	Reference: 
	"""
	def __init__(self, learner_func):
		"""
		Constructor.

		Parameters
		----------
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features.
			
			Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained.

			There should be as many importance scores in :code:`feature_importances_` as columns in :code:`x_train`.

		"""
		self.selected_variables = []
		self.learner_func = learner_func


	def fit(self, x_df, y_df, n_evaluations=20, pval=0.95, max_duration=None):
		"""
		Performs a run of the Boruta feature selection algorithm. 

		Specifically, :code:`n_evaluations` times we randomly shuffle all feature values to define so-called 'shadow features', which we add to the original features.

		The random shuffling destroys any association between shadow features and the target. Thus, none of them should be considered important.

		We train the learner with both original and shadow features, and calculate the highest feature importance score among shadow features. 

		We consider an original feature a 'hit' if its feature importance score is higher than the highest feature importance score of shadow features, otherwise it is a miss.

		For each original feature, we end up with the number of hits in the random :code:`n_evaluations` runs, and we use the following statistical test to conclude if the original feature is relevant.

		The null hypothesis :math:`H_0` is that we do not know if an original feature is relevant for the problem at hand.

		Under :math:`H_0`, the probability that a feature will be a hit in one of the random trials above is :math:`p=0.5`, and its number of hits after :code:`n_evaluations` runs follows a binomial distribution with probability :math:`0.5` and total number of trials :math:`n_evaluations`.

		We reject :math:`H_0` and accept the alternative hypothesis :math:`H_1` that a feature is relevant when the number of hits observed is higher than the :code:`pval` quantile under :math:`H_0`.

		We reject :math:`H_0` and accept the alternative hypothesis :math:`H_2` that a feature is *not* relevant when the number of hits observed is lower than the :code:`1-pval` quantile under :math:`H_0`.

		A warning is raised when we were not able to reject :math:`H_0` for some features.

		Features that were determined to be relevant are used to train the final learner, which this method returns. 

		When no feature was determined to be relevant, this method returns :code:`None`.


		Parameters
		----------
		x_df : pd.DataFrame
			A dataframe containing all features.
		y_df : pd.DataFrame
			A dataframe containing the target.
		n_evaluations : int (default 20)
			Number of trials in the Boruta algorithm.
		pval : float (between 0 and 1, default 0.95)
			Quantile level above which features are considered relevant.
		max_duration : float | None (default)
			If not None, then feature elimination will stop after this many seconds.


		Attributes
		----------
		selected_variables : list
			The list of features deemed relevant, sorted in decreasing number of hits.
		ambiguous_variables : list
			The list of features that we could not conclude are either relevant or irrelevant.
		hits : dict
			Dictionary containing the number of hits for each feature.


		Returns
		-------
		m : sklearn-like model (an instance returned by :code:`learner_func`)
			An instance returned by :code:`learner_func` trained with features that were deemed relevant.

		"""
		columns = [_ for _ in x_df.columns]
		y = y_df.values

		hits = {col: 0 for col in columns}
		start_time = time()
		for trial in tqdm(range(n_evaluations)):
			# Construct shaddow features
			train_x_df = pd.DataFrame(x_df.sample(frac=1).values, \
				index=x_df.index, columns=['shaddow_%s' % col for col in columns])

			# Add to real features
			train_x_df = pd.concat([x_df, train_x_df], axis=1)
			all_columns = [_ for _ in train_x_df.columns]

			# Fit the model
			x = train_x_df[all_columns].values
			current_n_vars = len(all_columns)
			m = self.learner_func(n_vars=current_n_vars)
			m.fit(x, y)

			# Increase the hit count
			importances = [_ for _ in m.feature_importances_]
			assert len(importances) == len(all_columns), \
				'The number of importance scores should be %d, received %d' % (len(all_columns), len(importances))
			hit_threshold = np.max([np.abs(importances[i]) \
				for i in range(current_n_vars) if all_columns[i].startswith('shaddow_')])

			for i in range(current_n_vars):
				if not all_columns[i].startswith('shaddow_') and np.abs(importances[i]) > hit_threshold:
					hits[all_columns[i]] = hits[all_columns[i]] + 1

			duration = time()-start_time
			if max_duration and duration > max_duration:
				logging.warning('We have exceeded the configured maximum duration %.2fs: exiting after %d trials...' % (max_duration, trial+1))
				n_evaluations = trial+1
				break


		# Testing whether the number of hits is statistically significant to conclude the feature is important.
		# H0: We can't tell if a feature is important or not. Said differently, there is 50% chance that a feature will be a 'hit' in a round.
		# H1: A feature is important.
		# H2: A feature is unimportant.

		# Under H0, the number of hits after n_evaluations trials should be in [hm, hM] with probability pval, where hm and hM are binomial quantiles.
		# We reject H0 and accept H1 when a feature's number of hits exceeds hM.
		# We reject H0 and accept H2 when a feature's number of hits is lower than hm. 
		cdfs = binom.cdf([hits[col] for col in columns], n_evaluations, 0.5)
		selected_variables = [columns[i] for i in range(len(columns)) if cdfs[i] > pval]
		associated_hits = [hits[col] for col in selected_variables]

		self.selected_variables = [col for _,  col in sorted(zip(associated_hits, selected_variables), reverse=True)]
		self.ambiguous_variables = [columns[i] for i in range(len(columns)) if cdfs[i] <= pval and cdfs[i] >= 1.-pval]

		if self.ambiguous_variables:
			logging.warning('Features %s have not been determined important or unimportant. Consider increasing n_evaluations.' % \
				str(self.ambiguous_variables))

		self.hits = hits
		if self.selected_variables == []:
			logging.warning('No variable/feature was deemed important by the model.')
			return None

		# Keep only important features and retrain the model.
		x = x_df[self.selected_variables].values
		n_vars = len(self.selected_variables)
		m = self.learner_func(n_vars=n_vars)
		m.fit(x, y)

		return m





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Estimation of the highest performance achievable in a supervised learning problem.
E.g. :math:`R^2`, RMSE, classification accuracy, true log-likelihood per observation.
"""
import logging
logging.basicConfig(level=logging.INFO)
import requests
import sys
from time import time, sleep

import numpy as np
import pandas as pd

try:
	get_ipython().__class__.__name__
	from halo import HaloNotebook as Halo
except:
	from halo import Halo

from kxy.api import APIClient, upload_data

# Cache old job ids to avoid being charged twice for the same job.
VALUATION_JOB_IDS = {}

def data_valuation(data_df, target_column, problem_type, snr='auto'):
	"""
	.. _data-valuation:
	Estimate the highest performance metrics achievable when predicting the :code:`target_column` using all other columns.

	When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`target_column` is categorical.


	Parameters
	----------
	data_df : pandas.DataFrame
		The pandas DataFrame containing the data.
	target_column : str
		The name of the column containing true labels.
	problem_type : None | 'classification' | 'regression'
		The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.



	Returns
	-------
	achievable_performance : pandas.Dataframe
		The result is a pandas.Dataframe with columns (where applicable):

		* :code:`'Achievable Accuracy'`: The highest classification accuracy that can be achieved by a model using provided inputs to predict the label.
		* :code:`'Achievable R-Squared'`: The highest :math:`R^2` that can be achieved by a model using provided inputs to predict the label.
		* :code:`'Achievable RMSE'`: The lowest Root Mean Square Error that can be achieved by a model using provided inputs to predict the label.		
		* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a model using provided inputs to predict the label.


	.. admonition:: Theoretical Foundation

		Section :ref:`1 - Achievable Performance`.
	"""
	assert target_column in data_df.columns, 'The label column should be a column of the dataframe.'
	assert problem_type.lower() in ['classification', 'regression']
	if problem_type.lower() == 'regression':
		assert np.can_cast(data_df[target_column], float), 'The target column should be numeric'

	k = 0
	max_k = 100

	file_name = upload_data(data_df)
	spinner = Halo(text='Waiting for results from the backend.', spinner='dots')
	spinner.start()

	if file_name:
		job_id = VALUATION_JOB_IDS.get((file_name, target_column, problem_type, snr), None)

		if job_id:
			api_response = APIClient.route(
				path='/wk/data-valuation', method='POST', 
				file_name=file_name, target_column=target_column, \
				problem_type=problem_type, \
				timestamp=int(time()), job_id=job_id, \
				snr=snr)
		else:
			api_response = APIClient.route(
				path='/wk/data-valuation', method='POST', \
				file_name=file_name, target_column=target_column, \
				problem_type=problem_type, timestamp=int(time()), \
				snr=snr)

		initial_time = time()
		while api_response.status_code == requests.codes.ok and k < max_k:
			try:
				response = api_response.json()
				if 'eta' in response:
					progress_text = '%s%% Completed.' % response['progress_pct'] if 'progress_pct' in response else ''
					spinner.text = 'Waiting for results from the backend. ETA: %s. %s' % (response['eta'], progress_text)

				if ('job_id' in response) and ('r-squared' not in response):
					job_id = response['job_id']
					VALUATION_JOB_IDS[(file_name, target_column, problem_type, snr)] = job_id
					k += 1
					sleep(15.)

					# Note: it is important to pass the job_id to avoid being charged twice for the same work.
					api_response = APIClient.route(
						path='/wk/data-valuation', method='POST', 
						file_name=file_name, target_column=target_column, \
						problem_type=problem_type, \
						timestamp=int(time()), job_id=job_id, \
						snr=snr)

					try:
						response = api_response.json()
						if 'eta' in response:
							progress_text = '%s%% Completed.' % response['progress_pct'] if 'progress_pct' in response else ''
							spinner.text = 'Waiting for results from the backend. ETA: %s. %s' % (response['eta'], progress_text)
					except:
						pass

				if ('job_id' not in response) or ('r-squared' in response):
					duration = int(time()-initial_time)
					duration = str(duration) + 's' if duration < 60 else str(duration//60) + 'min'

					result = {}
					if 'r-squared' in response:
						result['Achievable R-Squared'] = [response['r-squared']]

					if 'log-likelihood' in response:
						result['Achievable Log-Likelihood Per Sample'] = [response['log-likelihood']]

					if 'rmse' in response and problem_type.lower() == 'regression':
						result['Achievable RMSE'] = [response['rmse']]

					if 'accuracy' in response and problem_type.lower() == 'classification':
						result['Achievable Accuracy'] = [response['accuracy']]

					result = pd.DataFrame.from_dict(result)

					spinner.text = 'Received results from the backend after %s.' % duration
					spinner.succeed()

					return result

			except:
				logging.exception('\nData valuation failed. Last HTTP code: %s' % api_response.status_code)
				spinner.text = 'The backend encountered an unexpected error we are looking into. Please try again later.'
				spinner.fail()
				return None


		if api_response.status_code != requests.codes.ok:
			spinner.text = 'The backend is taking longer than expected. Try again later.'
			spinner.fail()
			try:
				response = api_response.json()
				if 'message' in response:
					logging.error('\n%s' % response['message'])
			except:
				logging.error('\nData valuation failed. Last HTTP code: %s' % api_response.status_code)

	return None





# -*- coding: utf-8 -*-
"""
Created on Thu Feb 27 10:54:00 2020

@author: ylkomsamo
"""

import sys
sys.path.append('.')
from setuptools import setup, find_packages

with open('README.md') as f:
	long_description = f.read()
	
version = "1.3.5"
setup(name="kxy",
	version=version,
	zip_safe=False,
	license="GPLv3",
	author="Dr. Yves-Laurent Kom Samo",
	author_email="github@kxy.ai",
	url="https://www.kxy.ai",
	description = "A Powerful Serverless Pre-Learning and Post-Learning Analysis Toolkit",
	long_description=long_description,
	long_description_content_type='text/markdown',  # This is important!
	project_urls={
		"Documentation": "https://www.kxy.ai/reference",
		"Source Code": "https://github.com/kxytechnologies/kxy-python/"},
	download_url = "https://github.com/kxytechnologies/kxy-python/archive/v%s.tar.gz" % version,
	keywords = ["Lean ML", "AutoML", "Pre-Learning", "Post-Learning", "Model-Free ML"],
	packages=find_packages(exclude=["tests"]),
	install_requires=["numpy>=1.13.1", "scipy>=1.4.1", "pandas>=0.23.0", "requests>=2.22.0", "pandarallel", "halo", "ipywidgets", "scikit-learn"],
	classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
		"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
		"Programming Language :: Python :: 3 :: Only",
		"Development Status :: 5 - Production/Stable",
		"Topic :: Scientific/Engineering :: Artificial Intelligence",
		"Topic :: Scientific/Engineering :: Mathematics"
	],
    scripts=['bin/kxy']
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from time import time
import numpy as np

class NaiveLearner(object):
	"""
	Implementation of the Recursive Feature Elimination (RFE) feature selection algorithm.

	Reference: 
	"""
	def __init__(self, learner_func):
		"""
		Constructor.

		Parameters
		----------
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features.
			
			Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained.

			There should be as many importance scores in :code:`feature_importances_` as columns in :code:`x_train`.

		"""
		self.selected_variables = []
		self.learner_func = learner_func


	def fit(self, x_df, y_df):
		"""
		Fit the model without feature selection.

		Parameters
		----------
		x_df : pd.DataFrame
			A dataframe containing all features.
		y_df : pd.DataFrame
			A dataframe containing the target.

		Attributes
		----------
		selected_variables : list
			The list of features.

		Returns
		-------
		m : sklearn-like model (an instance returned by :code:`learner_func`)
			An instance returned by :code:`learner_func` trained with all features.

		"""
		columns = [_ for _ in x_df.columns]
		y = y_df.values
		x = x_df[columns].values
		n_vars = len(columns)
		m = self.learner_func(n_vars=n_vars)
		m.fit(x, y)
		self.selected_variables = columns

		return m





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from time import time
import numpy as np
from tqdm import tqdm

class RFE(object):
	"""
	Implementation of the Recursive Feature Elimination (RFE) feature selection algorithm.

	Reference: 
	"""
	def __init__(self, learner_func):
		"""
		Constructor.

		Parameters
		----------
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features.
			
			Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained.

			There should be as many importance scores in :code:`feature_importances_` as columns in :code:`x_train`.

		"""
		self.selected_variables = []
		self.learner_func = learner_func


	def fit(self, x_df, y_df, n_vars, max_duration=None):
		"""
		Performs a run of the Recursive Feature Elimination (RFE) feature selection algorithm.

		Starting with all features, we recursively train a learner, calculate all feature importance scores, remove the least important feature, and repeat until we are left with :code:`n_vars` features.

		Parameters
		----------
		x_df : pd.DataFrame
			A dataframe containing all features.
		y_df : pd.DataFrame
			A dataframe containing the target.
		n_vars : int
			The number of features to keep.
		max_duration : float | None (default)
			If not None, then feature elimination will stop after this many seconds.

		Attributes
		----------
		selected_variables : list
			The list of the :code:`n_vars` features we kept.


		Returns
		-------
		m : sklearn-like model (an instance returned by :code:`learner_func`)
			An instance returned by :code:`learner_func` trained with the :code:`n_vars` features we kept.

		"""
		columns = [_ for _ in x_df.columns]
		y = y_df.values

		# Fit the model
		x = x_df[columns].values
		current_n_vars = len(columns)
		start_time = time()
		m = self.learner_func(n_vars=current_n_vars)
		m.fit(x, y)
		importances = [_ for _ in m.feature_importances_]

		n_rounds = max(current_n_vars-n_vars, 0)
		for _ in tqdm(range(n_rounds)):
			duration = time()-start_time
			if max_duration and duration > max_duration:
				logging.warning('We have exceeded the configured maximum duration %.2fs: exiting...' % max_duration)
				break

			# Remove the least important variable
			importances = [_ for _ in m.feature_importances_]
			least_important_ix = np.argmin(np.abs(importances))
			importances.pop(least_important_ix)
			least_important_feature = columns[least_important_ix]
			logging.info('Deleting feature %s' % least_important_feature)
			columns.remove(least_important_feature)
			current_n_vars = len(columns)

			# Re-fit the model
			x = x_df[columns].values
			m = self.learner_func(n_vars=current_n_vars)
			m.fit(x, y)

		self.selected_variables = [col for _,  col in sorted(zip(importances, columns), reverse=True)]

		return m





#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect
import gc
import logging
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, r2_score, mean_squared_error, roc_auc_score, \
	explained_variance_score

import pickle as pkl

try:
	get_ipython().__class__.__name__
	from halo import HaloNotebook as Halo
except:
	from halo import Halo

from ..pandas_extension.features_utils import rmspe_score, neg_rmspe_score, neg_mae_score, neg_rmse_score
from ..pandas_extension.pre_learning_accessor import PreLearningAccessor



class BaselineClassifier(object):
	"""
	Classifier following the scikit-learn API and always predicting the most frequent class in-sample.
	"""
	def fit(self, x, y):
		values, counts = np.unique(y, return_counts=True)
		ind = np.argmax(counts)
		self.training_mode = values[ind]

	def predict(self, x):
		try:
			n = x.shape[0]
			return np.array([self.training_mode]*n)[:, None]

		except AttributeError:
			logging.error('The model should be fitted first')


	def __getstate__(self):
		state = {'training_mode': self.training_mode}
		return state

	def save(self, path):
		with open(path, 'wb') as f:
			pkl.dump(self, f)

	def __setstate__(self, state):
		self.training_mode = state['training_mode']

	@classmethod
	def load(cls, path):
		with open(path, 'rb') as f:
			model = pkl.load(f)
		return model



class BaselineRegressor(object):
	"""
	Regressor following the scikit-learn API and always predicting the in-sample mean.
	"""
	def __init__(self, baseline='mean'):
		assert baseline in ['mean', 'median']
		self._baseline = np.nanmean if baseline == 'mean' else np.nanmedian


	def fit(self, x, y):
		self.training_baseline = self._baseline(y)

	def predict(self, x):
		try:
			n = x.shape[0]
			return np.ones((n, 1))*self.training_baseline

		except AttributeError:
			logging.error('The model should be fitted first')

	def __getstate__(self):
		state = {
			'baseline': 'mean' if self._baseline == np.mean else 'median', \
			'training_baseline': self.training_baseline
		}
		return state

	def save(self, path):
		with open(path, 'wb') as f:
			pkl.dump(self, f)

	def __setstate__(self, state):
		self.training_baseline = state['training_baseline']
		self._baseline = np.nanmean if state['baseline'] == 'mean' else np.nanmedian

	@classmethod
	def load(cls, path):
		with open(path, 'rb') as f:
			model = pkl.load(f)
		return model




class ShrunkLearner(object):
	"""
	Wrapper to seamlessly add effective variable selection to any supervised learner. 
	"""
	def _non_additive_fit(self, obj, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=3, \
			regression_baseline='mean', regression_error_type='additive', return_scores=False, start_n_features_perf_frac=0.9, \
			val_performance_buffer=0.0):
		# A base learner here does not fix mistakes made by another.
		assert inspect.isfunction(learner_func), 'learner_func should be a class'
		assert target_column in obj.columns, 'The target column should be a valid column'
		if problem_type is None:
			problem_type = 'classification' if obj.kxy.is_discrete(target_column) else 'regression'
		assert problem_type in ('classification', 'regression')
		self.problem_type = problem_type
		self.additive_learning = False
		assert regression_error_type in ('additive', 'multiplicative')
		self.regression_error_type = regression_error_type

		for col in obj.columns:
			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'

		x_columns = [_ for _ in obj.columns if _ != target_column]
		if self.problem_type == 'classification':
			labels = set(list(obj[target_column].values.astype(int)))
			binary_labels = {0, 1}
			assert labels.issubset(binary_labels), 'Classification labels should either be 0 or 1'

		if benchmark_feature:
			assert benchmark_feature in obj.columns, 'The benchmark feature should be a valid column'
		self.benchmark_feature = benchmark_feature
		if callable(score):
			score_func = score
		else:
			if score == 'auto':
				score = 'r2_score' if problem_type == 'regression' else 'accuracy_score'
			score_func = eval(score)
		score_name = score_func.__name__

		self.val_scores = []

		if getattr(self, 'models', None) is None or force_redo:
			# 0. Train/Validation split
			self.target_column = target_column
			if return_scores:
				# Reserve 1-train_frac for testing, and train_frac for training and validation
				self.test_df = obj.sample(frac=1.-train_frac, random_state=random_state)
				self.train_val_df = obj.drop(self.test_df.index)
			else:
				self.train_val_df = obj
			# Reserve train_frac*train_frac for training [...]
			self.train_df = self.train_val_df.sample(frac=train_frac, random_state=random_state)
			# [...] and train_frac*(1-train_frac) for validation.
			self.val_df = self.train_val_df.drop(self.train_df.index)

			if missing_value_imputation:
				# Basic missing value imputation
				self.train_df.fillna(self.train_df.median(), inplace=True)
				self.val_df.fillna(self.train_df.median(), inplace=True)
				if return_scores:
					self.test_df.fillna(self.train_df.median(), inplace=True)

			# 1. Model-free variable selection
			vs_accessor = PreLearningAccessor(obj)
			self.variable_selection_results = vs_accessor.variable_selection(self.target_column, problem_type=self.problem_type, \
				snr=snr, anonymize=anonymize)

			self.variables = [_ for _ in self.variable_selection_results['Variable'].values if _.lower() != 'no variable']
			n_variables = len(self.variables)
			if max_n_features:
				n_variables = min(n_variables, max_n_features)

			if start_n_features is None:
				perfs = [_ for _ in self.variable_selection_results['Running Achievable R-Squared'].astype(float)]
				max_perf = np.max(perfs)
				perf_threshold = start_n_features_perf_frac*max_perf
				start_n_features = n_variables-len([_ for _ in perfs if _ > perf_threshold])+1
				logging.info('Starting with %d variables' % start_n_features)

			# 2. Sequentially add variables in decreasing order of importance.
			# 2.1 Baseline performance
			y_train = self.train_df[[self.target_column]].values
			y_val = self.val_df[[self.target_column]].values
			if return_scores:
				y_test = self.test_df[[self.target_column]].values
			x_train = self.train_df[self.variables[:1]].values
			x_val = self.val_df[self.variables[:1]].values

			base_m = BaselineRegressor(baseline=regression_baseline) if problem_type == 'regression' else BaselineClassifier()
			base_m.fit(x_train, y_train)
			y_val_pred = base_m.predict(x_val)
			previous_score = score_func(y_val, y_val_pred)

			spinner = Halo(text='Lean Boosting:', spinner='dots')
			spinner.start()
			logging.info('Baseline score (%s): %.4f' % (score_name, previous_score))
			spinner.text = 'Lean Boosting -- Baseline %s: %.4f' % (score_name, previous_score)

			self.start_n_features = min(start_n_features, n_variables)
			self.models = []
			self.max_var_ixs = []
			n_down_perf = 0
			for i in range(self.start_n_features, n_variables+1):
				gc.collect()
				vs = self.variables[:i]
				x_train = self.train_df[vs].values
				x_val = self.val_df[vs].values
				n_vars = x_train.shape[1] if len(x_train.shape) > 1 else 1

				# Create the new model
				m = learner_func(n_vars=n_vars)

				# Fit the new model
				m.fit(x_train, y_train)

				# New validation score
				y_val_pred = m.predict(x_val)
				val_score = score_func(y_val, y_val_pred)

				if val_score > previous_score+val_performance_buffer or (min_n_features and i<=min_n_features):
					n_down_perf = 0
					logging.info('Variable #%d (%s) increased validation performance from %.4f to %.4f' % (i, self.variables[i-1], previous_score, val_score))
					spinner.text = 'Lean Boosting -- %d Variables, Validation %s: %.4f' % (i, score_name, val_score)
					previous_score = val_score
					self.models = [m]
					self.max_var_ixs = [i]
					self.val_scores = self.val_scores + [(i, val_score)]

				else:
					n_down_perf += 1
					logging.info('Validation performance did not increase for the %d-th consecutive time. Old: %.4f, New: %.4f, Variable: %s' % (n_down_perf, previous_score, val_score, self.variables[i-1]))
					if n_down_perf >= n_down_perf_before_stop:
						# Only stop after a certain number of consecutive down performance
						logging.info('Stopping training as validation performance did not increase %d consecutive times.' % n_down_perf_before_stop)
						spinner.succeed()
						break

				if max_n_features and (i==max_n_features):
					logging.info('Stopping training as the maximum number of variables (%d) has been reached' % max_n_features)
					spinner.succeed()
					break		

			self.selected_variables = self.variables[:self.max_var_ixs[-1]] if self.models else []
			if self.models == []:
				self.models = [base_m]
				self.max_var_ixs = [1]
				self.val_scores = [(0, previous_score)]

			results = {'Selected Variables': self.selected_variables}
			if return_scores:
				# Inputs
				x_train = self.train_df[self.selected_variables].values
				x_val = self.val_df[self.selected_variables].values
				x_test = self.test_df[self.selected_variables].values

				# Predictions
				self.y_train_pred = self.models[0].predict(x_train)
				self.y_val_pred = self.models[0].predict(x_val)
				self.y_test_pred = self.models[0].predict(x_test)

				# Scores
				self.train_score = score_func(y_train, self.y_train_pred)
				self.val_score = score_func(y_val, self.y_val_pred)
				self.test_score = score_func(y_test, self.y_test_pred)

				results['Training Score'] = '%.5f' % self.train_score
				results['Validation Score'] = '%.5f' % self.val_score
				results['Testing Score'] = '%.5f' % self.test_score

				if self.problem_type == 'regression':
					results['Training R-Squared'] = '%.3f' % r2_score(y_train.flatten(), self.y_train_pred.flatten())
					results['Validation R-Squared'] = '%.3f' % r2_score(y_val.flatten(), self.y_val_pred.flatten())
					results['Testing R-Squared'] = '%.3f' % r2_score(y_test.flatten(), self.y_test_pred.flatten())

					results['Training RMSE'] = '%.5f' % mean_squared_error(y_train.flatten(), self.y_train_pred.flatten(), squared=False)
					results['Validation RMSE'] = '%.5f' % mean_squared_error(y_val.flatten(), self.y_val_pred.flatten(), squared=False)
					results['Testing RMSE'] = '%.5f' % mean_squared_error(y_test.flatten(), self.y_test_pred.flatten(), squared=False)

				if self.problem_type == 'classification':
					results['Training Accuracy'] = '%.3f' % self.train_score
					results['Validation Accuracy'] =  '%.3f' % self.val_score
					results['Testing Accuracy'] = '%.3f' % self.test_score

			return results




	def _additive_fit(self, obj, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=3, \
			regression_baseline='mean', regression_error_type='additive', return_scores=False, start_n_features_perf_frac=0.9, \
			val_performance_buffer=0.0):
		# A base learner here is fitted to the residuals of the best model so far.
		assert inspect.isfunction(learner_func), 'learner_func should be a class'
		assert target_column in obj.columns, 'The target column should be a valid column'
		if problem_type is None:
			problem_type = 'classification' if obj.kxy.is_discrete(target_column) else 'regression'
		assert problem_type in ('classification', 'regression')
		self.problem_type = problem_type
		self.additive_learning = True
		assert regression_error_type in ('additive', 'multiplicative')
		self.regression_error_type = regression_error_type

		for col in obj.columns:
			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'

		x_columns = [_ for _ in obj.columns if _ != target_column]
		if self.problem_type == 'classification':
			labels = set(list(obj[target_column].values.astype(int)))
			binary_labels = {0, 1}
			assert labels.issubset(binary_labels), 'Classification labels should either be 0 or 1'

		if benchmark_feature:
			assert benchmark_feature in obj.columns, 'The benchmark feature should be a valid column'
		self.benchmark_feature = benchmark_feature
		if callable(score):
			score_func = score
		else:
			if score == 'auto':
				score = 'r2_score' if problem_type == 'regression' else 'accuracy_score'
			score_func = eval(score)
		score_name = score_func.__name__
		self.val_scores = []

		if getattr(self, 'models', None) is None or force_redo:
			# 0. Train/Validation split
			self.target_column = target_column
			if return_scores:
				# Reserve 1-train_frac for testing, and train_frac for training and validation
				self.test_df = obj.sample(frac=1.-train_frac, random_state=random_state)
				self.train_val_df = obj.drop(self.test_df.index)
			else:
				self.train_val_df = obj
			# Reserve train_frac*train_frac for training [...]
			self.train_df = self.train_val_df.sample(frac=train_frac, random_state=random_state)
			# [...] and train_frac*(1-train_frac) for validation.
			self.val_df = self.train_val_df.drop(self.train_df.index)

			if missing_value_imputation:
				# Basic missing value imputation
				self.train_df.fillna(self.train_df.median(), inplace=True)
				self.val_df.fillna(self.train_df.median(), inplace=True)
				if return_scores:
					self.test_df.fillna(self.train_df.median(), inplace=True)

			# 1. Model-free variable selection
			vs_accessor = PreLearningAccessor(obj)
			self.variable_selection_results = vs_accessor.variable_selection(self.target_column, problem_type=self.problem_type, \
				snr=snr, anonymize=anonymize)

			self.variables = [_ for _ in self.variable_selection_results['Variable'].values if _.lower() != 'no variable']
			n_variables = len(self.variables)
			if max_n_features:
				n_variables = min(n_variables, max_n_features)

			if start_n_features is None:
				perfs = [_ for _ in self.variable_selection_results['Running Achievable R-Squared'].astype(float)]
				max_perf = np.max(perfs)
				perf_threshold = start_n_features_perf_frac*max_perf
				start_n_features = n_variables-len([_ for _ in perfs if _ > perf_threshold])+1

			# 2. Sequentially add variables in decreasing order of importance.
			# 2.1 Baseline performance
			y_train = self.train_df[[self.target_column]].values
			y_val = self.val_df[[self.target_column]].values
			if return_scores:
				y_test = self.test_df[[self.target_column]].values
			x_train = self.train_df[self.variables[:1]].values
			x_val = self.val_df[self.variables[:1]].values

			base_m = BaselineRegressor(baseline=regression_baseline) if problem_type == 'regression' else BaselineClassifier()
			base_m.fit(x_train, y_train)
			y_val_pred = base_m.predict(x_val)
			previous_score = score_func(y_val, y_val_pred)

			spinner = Halo(text='Lean Boosting:', spinner='dots')
			spinner.start()
			logging.info('Baseline score (%s): %.4f' % (score_name, previous_score))
			spinner.text = 'Lean Boosting -- Baseline %s: %.4f' % (score_name, previous_score)

			self.start_n_features = min(start_n_features, n_variables)
			n_down_perf = 0
			target_train = y_train.copy()
			y_val_pred = None
			self.models = []
			self.max_var_ixs = []
			for i in range(self.start_n_features, n_variables+1):
				gc.collect()
				vs = self.variables[:i]
				x_train = self.train_df[vs].values
				x_val = self.val_df[vs].values
				n_vars = x_train.shape[1] if len(x_train.shape) > 1 else 1

				# Create the new model
				m = learner_func(n_vars=n_vars)

				# Fit the new model
				m.fit(x_train, target_train)

				# New validation score
				target_val_pred = m.predict(x_val)
				target_val_pred = target_val_pred if len(target_val_pred.shape) > 1 else target_val_pred[:, None]
				target_train_pred = m.predict(x_train)
				target_train_pred = target_train_pred if len(target_train_pred.shape) > 1 else target_train_pred[:, None]

				if y_val_pred is None:
					y_val_pred = target_val_pred
				else:
					if self.problem_type == 'regression':
						y_val_pred = y_val_pred+target_val_pred if self.regression_error_type == 'additive' else \
							y_val_pred*target_val_pred

					if self.problem_type == 'classification':
						y_val_pred = np.abs(y_val_pred-target_val_pred)

				val_score = score_func(y_val, y_val_pred)
				if val_score > previous_score+val_performance_buffer or (min_n_features and i<=min_n_features):
					n_down_perf = 0
					logging.info('Variable #%d (%s) increased validation performance from %.4f to %.4f' % (i, self.variables[i-1], previous_score, val_score))
					spinner.text = 'Lean Boosting -- %d Variables, Validation %s: %.4f' % (i, score_name, val_score)
					previous_score = val_score

					if self.problem_type == 'regression':
						target_train = target_train-target_train_pred if self.regression_error_type == 'additive' else target_train/target_train_pred

					if self.problem_type == 'classification':
						target_train = np.logical_not(target_train == target_train_pred).astype(int)

					self.models = self.models+[m]
					self.max_var_ixs = self.max_var_ixs+[i]
					self.val_scores = self.val_scores + [(i, val_score)]

				else:
					n_down_perf += 1
					logging.info('Validation performance did not increase for the %d-th consecutive time. Old: %.4f, New: %.4f, Variable: %s' % (n_down_perf, previous_score, val_score, self.variables[i-1]))
					if n_down_perf >= n_down_perf_before_stop:
						# Only stop after a certain number of consecutive down performance
						logging.info('Stopping training as validation performance did not increase %d consecutive times.' % n_down_perf_before_stop)
						spinner.succeed()
						break

				if max_n_features and (i==max_n_features):
					logging.info('Stopping training as the maximum number of variables (%d) has been reached' % max_n_features)
					spinner.succeed()
					break	

			self.selected_variables = self.variables[:self.max_var_ixs[-1]] if self.models else []
			if self.models == []:
				self.models = [base_m]
				self.max_var_ixs = [1]
				self.val_scores = [(0, previous_score)]

			results = {'Selected Variables': self.selected_variables}
			if return_scores:
				# Inputs
				x_train = self.train_df[self.selected_variables].values
				x_val = self.val_df[self.selected_variables].values
				x_test = self.test_df[self.selected_variables].values

				# Predictions
				self.y_train_pred = self.predict(self.train_df)
				self.y_train_pred = self.y_train_pred.values.flatten()

				self.y_val_pred = self.predict(self.val_df)
				self.y_val_pred = self.y_val_pred.values.flatten()

				self.y_test_pred = self.predict(self.test_df)
				self.y_test_pred = self.y_test_pred.values.flatten()

				# Scores
				self.train_score = score_func(y_train, self.y_train_pred)
				self.val_score = score_func(y_val, self.y_val_pred)
				self.test_score = score_func(y_test, self.y_test_pred)

				results['Training Score'] = '%.5f' % self.train_score
				results['Validation Score'] = '%.5f' % self.val_score
				results['Testing Score'] = '%.5f' % self.test_score

				if self.problem_type == 'regression':
					results['Training R-Squared'] = '%.3f' % r2_score(y_train.flatten(), self.y_train_pred.flatten())
					results['Validation R-Squared'] = '%.3f' % r2_score(y_val.flatten(), self.y_val_pred.flatten())
					results['Testing R-Squared'] = '%.3f' % r2_score(y_test.flatten(), self.y_test_pred.flatten())

					results['Training RMSE'] = '%.5f' % mean_squared_error(y_train.flatten(), self.y_train_pred.flatten(), squared=False)
					results['Validation RMSE'] = '%.5f' % mean_squared_error(y_val.flatten(), self.y_val_pred.flatten(), squared=False)
					results['Testing RMSE'] = '%.5f' % mean_squared_error(y_test.flatten(), self.y_test_pred.flatten(), squared=False)

				if self.problem_type == 'classification':
					results['Training Accuracy'] = '%.3f' % self.train_score
					results['Validation Accuracy'] =  '%.3f' % self.val_score
					results['Testing Accuracy'] = '%.3f' % self.test_score

			return results



	def fit(self, obj, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=None, anonymize=False, \
			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=3, \
			regression_baseline='mean', additive_learning=False, regression_error_type='additive', return_scores=False, \
			start_n_features_perf_frac=0.9, val_performance_buffer=0.0):
		"""
		Train a lean boosted supervised learner, bringing in variables one at a time, in decreasing order of importance (as per :code:`obj.kxy.variable_selection`), until doing so no longer improves validation performance or another stopping criterion is met.

		Specifically, training proceeds as follows. First, KXY's model-free variable selection is run (i.e. :code:`obj.kxy.variable_selection`). 

		Then we train a model (instance returned by :code:`learner_func`) using the :code:`start_n_features` most important feature/variable to predict the target (defined by :code:`target_column`).

		Next we consider adding one variable at a time to fix the mistakes made by the previously trained model when :code:`additive_learning` is :code:`True`.
	
		If doing so improves performance on the validation set, we keep going until either performance no longer improves on the validation set :code:`n_down_perf_before_stop` consecutive times, or we've selected :code:`max_n_features` features.

		When :code:`start_n_features` is :code:`None` the initial set of variables is the smallest set of variables with which we may achieve :code:`start_n_features_perf_frac` of the performance we could achieve using all variables (as per :code:`obj.kxy.variable_selection`).

		When :code:`additive_learning` is set to :code:`False`, after adding a new variable, we will train the new model on the original problem, rather than trying to improve residuals.


		Parameters
		----------
		obj : pd.DataFrame
			The pandas dataframe containing training data.
		target_column : str
			The name of the column containing true labels.
		learner_func : function
			A function returning an instance of a base learner. They should define a :code:`fit(x, y)` method and a :code:`predict(x)` method.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		snr : 'auto' | 'low' | 'high'
			Set to :code:`low` if the problem is difficult (i.e. has a low signal-to-noise ratio) or the number of rows is small relative to the number of columns. Only used for model-free variable selection.
		train_frac : float
			The fraction of rows used for training and validation.
		random_state : int
			The seed to use for random training/validation/testing split.
		force_redo : bool
			Fitted models are saved. Set this parameters to :code:`True` to ignore saved models and refit.
		min_n_features : int | None
			Boosting will not stop until at least this many features/explanatory variables are selected.
		max_n_features : int | None
			Boosting will stop as soon as this many features/explanatory variables are selected.
		start_n_features : int
			The number of most important features boosting will start with.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).
		benchmark_feature : str | None
			When not None, 'benchmark' performance metrics using this column as predictor will be reported in the output dictionary.
		missing_value_imputation : bool
			When set to True, replace missing values with medians.
		n_down_perf_before_stop : int
			Number of consecutive down performances to observe before boosting stops.
		regression_baseline : str (:code:`mean` | :code:`median`)
			Whether to use the unconditional mean or median as the best predictor in the absence of explanatory variables.
			Choosing the mean corresponds to minimizing the L2 norm, whereas choosing the median corresponds to minimizing the L1 norm.
		additive_learning : bool
			When a new variable is added, whether errors/residuals should be fixed or a new model should be learned from scratch.
		regression_error_type : str ('additive' | 'multiplicative')
			For regression problems with additive learning, this determines whether the final model should be additive (pruning tries to reduce regressor residuals) or multiplicative (i.e. pruning tries to bring the ratio between true and predicted labels as closed to 1 as possible).
		start_n_features_perf_frac : float (between 0 and 1)
			When :code:`start_n_features` is not specified, it is set to the number of variables required to achieve a fraction :code:`start_n_features_perf_frac` of the maximum performance achievable (as per :code:`obj.kxy.variable_selection`).
		return_scores : bool (Default False)
			Whether to return training, validation and testing performance after lean boosting.
		val_performance_buffer : float (Default 0.0)
			The threshold by which the new validation performance needs to exceed the previously evaluated validation performance to consider increasing the number of features.
		score : str | func
			The metric to use to determine if a new feature should be added.




		Returns
		-------
		result : dict
			Dictionary containing selected variables, as well as training, validation and testing performance.
		"""

		if additive_learning:
			return self._additive_fit(obj, target_column, learner_func, problem_type=problem_type, snr=snr, train_frac=train_frac, random_state=random_state, \
				force_redo=force_redo, max_n_features=max_n_features, min_n_features=min_n_features, start_n_features=start_n_features, anonymize=anonymize, \
				benchmark_feature=benchmark_feature, missing_value_imputation=missing_value_imputation, score=score, n_down_perf_before_stop=n_down_perf_before_stop, \
				regression_baseline=regression_baseline, regression_error_type=regression_error_type, return_scores=return_scores, start_n_features_perf_frac=start_n_features_perf_frac, \
				val_performance_buffer=val_performance_buffer)

		else:
			return self._non_additive_fit(obj, target_column, learner_func, problem_type=problem_type, snr=snr, train_frac=train_frac, random_state=random_state, \
				force_redo=force_redo, max_n_features=max_n_features, min_n_features=min_n_features, start_n_features=start_n_features, anonymize=anonymize, \
				benchmark_feature=benchmark_feature, missing_value_imputation=missing_value_imputation, score=score, n_down_perf_before_stop=n_down_perf_before_stop, \
				regression_baseline=regression_baseline, regression_error_type=regression_error_type, return_scores=return_scores, start_n_features_perf_frac=start_n_features_perf_frac, \
				val_performance_buffer=val_performance_buffer)



	def _additive_predict(self, obj):
		# Global predict in an additive setting (when base learners are fitted to residuals, not the original problem).
		assert hasattr(self, 'models'), 'The model should be first fitted'
		for col in obj.columns:
			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'

		data = obj[self.variables]
		y_pred = None
		for i in range(len(self.models)):
			vs = self.variables[:self.max_var_ixs[i]]
			x = data[vs].values
			y_error_pred = self.models[i].predict(x)
			y_error_pred = y_error_pred if len(y_error_pred.shape) > 1 else y_error_pred[:, None]
			if y_pred is None:
				y_pred = y_error_pred
			else:
				if self.problem_type == 'regression':
					y_pred = y_pred+y_error_pred if self.regression_error_type == 'additive' else \
						y_pred*y_error_pred

				if self.problem_type == 'classification':
					y_pred = np.abs(y_pred-y_error_pred)

		y_pred = y_pred if len(y_pred.shape) > 1 else y_pred[:, None]
		predictions = pd.DataFrame(y_pred, columns=[self.target_column], index=data.index)

		return predictions


	def _non_additive_predict(self, obj):
		# Global predict in a non-additive setting (when base learners are fitted to the original problem, no residuals).
		assert hasattr(self, 'models'), 'The model should be first fitted'
		for col in obj.columns:
			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'

		data = obj[self.selected_variables]
		x = data.values
		y_pred = self.models[0].predict(x)
		y_pred = y_pred if len(y_pred.shape) > 1 else y_pred[:, None]
		predictions = pd.DataFrame(y_pred, columns=[self.target_column], index=data.index)

		return predictions


	def _predict(self, obj):
		assert hasattr(self, 'models'), 'The model should be first fitted'
		if self.additive_learning:
			return self._additive_predict(obj)
		else:
			return self._non_additive_predict(obj)


	def predict(self, obj, memory_bound=False):
		"""
		Make predictions using the fitted model.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing test explanatory variables about which we want to make predictions.
		memory_bound : bool (Default False)
			Whether we should try to save memory.


		Returns
		-------
		result : pandas.DataFrame
			A dataframe with the same index as :code:`obj`, and with one column whose name is the :code:`target_column` used for training.
		"""
		if memory_bound:
			n = obj.shape[0]
			max_n = 1000000
			res = pd.DataFrame(index=obj.index)
			res[self.target_column] = np.nan
			i = 0
			while i < n:
				res.iloc[i:i+max_n] = self._predict(obj.iloc[i:i+max_n])
				i += max_n
			return res

		else:
			return self._predict(obj)


	def save(self, path):
		""" """
		n_models = len(self.models)
		meta_path = path + '-meta-ShrunkLearner'
		meta = {\
			'additive_learning': self.additive_learning, 'target_column': self.target_column, 'variables': self.variables, \
			'selected_variables': self.selected_variables, 'problem_type': self.problem_type, 'max_var_ixs': self.max_var_ixs, \
			'n_models': n_models \
		}
		with open(meta_path, 'wb') as f:
			pkl.dump(meta, f)

		for i in range(n_models):
			self.models[i].save(path + '-model-%d-ShrunkLearner' % i)


	@classmethod
	def load(self, path, learner_func):
		""" 
		Load the learner from disk.
		"""
		meta_path = path + '-meta-ShrunkLearner'
		with open(meta_path, 'rb') as f:
			meta = pkl.load(f)

		additive_learning = meta['additive_learning']
		target_column = meta['target_column']
		variables = meta['variables']
		selected_variables = meta['selected_variables']
		problem_type = meta['problem_type']
		max_var_ixs = meta['max_var_ixs']
		n_models = meta['n_models']
		n_vars = len(selected_variables)

		predictor = ShrunkLearner()
		predictor.additive_learning = additive_learning
		predictor.target_column = target_column
		predictor.variables = variables
		predictor.selected_variables = selected_variables
		predictor.problem_type = problem_type
		predictor.max_var_ixs = max_var_ixs

		predictor.models = []
		for i in range(n_models):
			model_path = path + '-model-%d-ShrunkLearner' % i
			model = learner_func(n_vars=n_vars, path=model_path)
			predictor.models.append(model)

		return predictor



